import pytest

import numpy as np
from persim import PersImage

def test_landscape():
    bds = np.array([[1,1],[1,2]])

    ldsp = PersImage.to_landscape(bds)

    np.testing.assert_array_equal(ldsp, [[1,0],[1,1]])

class TestWeighting:
    def test_zero_on_xaxis(self):
        pim = PersImage()

        wf = pim.weighting()

        assert wf([1,0]) == 0
        assert wf([100,0]) == 0
        assert wf([99, 1.4]) == 1.4

    def test_scales(self):
        pim = PersImage()

        wf = pim.weighting(np.array([[0,1],[1,2],[3,4]]))

        assert wf([1,0]) == 0
        assert wf([1,4]) == 1
        assert wf([1,2]) == .5
    

class TestKernels:
    def test_kernel_mean(self):
        pim = PersImage()
        kf = pim.kernel(2)

        data = np.array([[0,0]])
        assert kf(np.array([[0,0]]), [0,0]) >= kf(np.array([[1,1]]), [0,0]), "decreasing away"
        assert kf(np.array([[0,0]]), [1,1]) == kf(np.array([[1,1]]), [0,0]), "symmetric"


class TestTransforms:
    def test_n_pixels(self):
        pim = PersImage(pixels=9)
        diagram = np.array([[0,1], [1,1],[3,5]])
        img = pim.transform(diagram)

        assert img.shape == (3,3)

    def test_multiple_diagrams(self):
        pim = PersImage(pixels=9)
        
        diagram1 = np.array([[0,1], [1,1],[3,5]])
        diagram2 = np.array([[0,1], [1,1],[3,6]])
        imgs = pim.transform([diagram1, diagram2])

        assert len(imgs) == 2
        assert imgs[0].shape == imgs[1].shape

@pytest.mark.skip()
class TestIntegration:
    """ We can't just take the center point, we need to integrate over the surface.

        It will be changing, so we need to ensure it works correctly.
    """
    def test_integrate_constant(self):
        intr = Integrator()
        
        f = lambda center:  1
        assert np.allclose(intr.integrate(f, [0], 2), 1 * (2*2)**2)
        
        f = lambda center:  2
        assert np.allclose(intr.integrate(f, [0], 2), 2 * (2*2)**2)
        
        f = lambda center:  3
        assert np.allclose(intr.integrate(f, [0], 2), 3 * (2*2)**2)

    def test_integrate_(self):
        # I am not sure these are correct
        intr = Integrator()
        
        f = lambda center:  center[0]
        
        assert np.allclose(intr.integrate(f, [0,10], 2), 32)
        
        f = lambda center:  center[0]
        assert np.allclose(intr.integrate(f, [1,10], 2), 32)
        
        f = lambda center:  center[0]
        assert np.allclose(intr.integrate(f, [2, 10], 2), 32)

    def test_coplanar(self):
        intr = Integrator()
        cube = np.array([[0,0,0],[1,0,0],[0,1,0],[1,1,0]])

 
        vol = intr._convex_hull_volume_bis(cube)
        assert np.allclose(vol, 0)

    def test_convex_hull_vol(self):
        intr = Integrator()

        cube = np.array([[0,0,0],[1,0,0],[0,1,0],[1,1,0],
                           [0,0,1],[1,0,1],[0,1,1],[1,1,1]])

        vol = intr._convex_hull_volume_bis(cube)
        assert np.allclose(vol, 1)

        vol = intr._convex_hull_volume_bis(cube*2)
        assert np.allclose(vol, 8)



from itertools import product

import numpy as np
from scipy.stats import multivariate_normal as mvn
from scipy.stats import norm
import scipy.spatial as spatial
import matplotlib.pyplot as plt

from sklearn.base import BaseEstimator


class PersImage(BaseEstimator):
    def __init__(
        self,
        pixels=20 * 20,
        spread=None,
        specs=None,
        kernel_type="gaussian",
        weighting_type="linear",
    ):
        """ Initialize a persistence image generator.

            pixels

        """

        self.specs = specs
        self.kernel_type = kernel_type
        self.weighting_type = weighting_type
        self.spread = spread

        self.pixels = pixels
        self.N = int(np.sqrt(pixels))

        print(
            'PersImage(pixels={}, spread={}, specs={}, kernel_type="{}", weighting_type="{}")'.format(
                pixels, spread, specs, kernel_type, weighting_type
            )
        )

    def transform(self, diagrams):
        """ Convert diagram or list of diagrams to a persistence image
        """

        # TODO: allow for nonsquare images
        assert int(np.sqrt(self.pixels)) == np.sqrt(
            self.pixels
        ), "Pixels must be a square"

        if type(diagrams) is not list:
            dg = np.copy(diagrams)  # keep original diagram untouched
            landscape = PersImage.to_landscape(dg)

            if not self.specs:
                self.specs = {"maxBD": np.max(landscape), "minBD": np.min(landscape)}

            imgs = self._transform(landscape)
        else:
            dgs = [np.copy(diagram) for diagram in diagrams]
            landscapes = [PersImage.to_landscape(dg) for dg in dgs]

            if not self.specs:
                self.specs = {
                    "maxBD": np.max([np.max(landscape) for landscape in landscapes]),
                    "minBD": np.min([np.min(landscape) for landscape in landscapes]),
                }
            imgs = [self._transform(dgm) for dgm in landscapes]

        return imgs

    def _transform(self, landscape):
        """ Convert single diagram to a persistence image
        """

        N = self.N

        # Define an NxN grid over our landscape
        maxBD = self.specs["maxBD"]
        minBD = min(self.specs["minBD"], 0)  # at least show 0, maybe lower

        # Same bins in x and y axis
        dx = maxBD / (N)
        xs_lower = np.linspace(minBD, maxBD, N)
        xs_upper = np.linspace(minBD, maxBD, N) + dx

        ys_lower = np.linspace(0, maxBD, N)
        ys_upper = np.linspace(0, maxBD, N) + dx

        weighting = self.weighting(landscape)

        # Define zeros
        img = np.zeros((N, N))

        # Implement this as a `summed-area table` - it'll be way faster
        spread = self.spread if self.spread else dx
        for point in landscape:
            x_smooth = norm.cdf(xs_upper, point[0], spread) - norm.cdf(
                xs_lower, point[0], spread
            )
            y_smooth = norm.cdf(ys_upper, point[1], spread) - norm.cdf(
                ys_lower, point[1], spread
            )
            img += np.outer(x_smooth, y_smooth) * weighting(point)
        img = img.T[::-1]
        return img

    def weighting(self, landscape=None):
        """ Define a weighting function, 
                for stability results to hold, the function must be 0 at y=0.    
        """

        # TODO: Implement a logistic function
        # TODO: use self.weighting_type to choose function

        if landscape is not None:
            maxy = np.max(landscape[:, 1])

        def linear(interval):
            # linear function of y such that f(0) = 0 and f(max(y)) = 1
            d = interval[1]
            return (1 / maxy) * d if landscape is not None else d

        def pw_linear(interval):
            """ This is the function defined as w_b(t) in the original PI paper

                Take b to be maxy/self.N to effectively zero out the bottom pixel row
            """

            t = interval[1]
            b = maxy / self.N

            if t <= 0:
                return 0
            if 0 < t < b:
                return t / b
            if b <= t:
                return 1

        return linear

    def kernel(self, spread=1):
        """ This will return whatever kind of kernel we want to use.
            Must have signature (ndarray size NxM, ndarray size 1xM) -> ndarray size Nx1
        """
        # TODO: use self.kernel_type to choose function

        def gaussian(data, pixel):
            return mvn.pdf(data, mean=pixel, cov=spread)

        return gaussian

    @staticmethod
    def to_landscape(diagram):
        """ Convert a diagram to a landscape
            (b,d) -> (b, d-b)
        """
        diagram[:, 1] -= diagram[:, 0]

        return diagram

    def show(self, imgs):
        """ Visualize the persistence image
        """

        if type(imgs) is not list:
            imgs = [imgs]

        for i, img in enumerate(imgs):
            plt.imshow(img, cmap=plt.get_cmap("plasma"))
            # plt.show()



#!/usr/bin/env python

from setuptools import setup

with open('README.md') as f:
    long_description = f.read()

setup(name='persim',
      version='0.0.6',
      description='Python implementation persistent images representation of persistence diagrams.',
      long_description=long_description,
      long_description_content_type="text/markdown",	
      author='Nathaniel Saul',
      author_email='nathaniel.saul@wsu.edu',
      url='https://github.com/sauln/persim',
      license='MIT',
      packages=['persim'],
      include_package_data=True,
      install_requires=[
        'scikit-learn',
        'numpy',
        'matplotlib',
        'scipy'
      ],
      python_requires='>=2.7,!=3.1,!=3.2,!=3.3',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: Healthcare Industry',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
      ],
      keywords='persistent homology, persistence images, persistence diagrams, topology data analysis, algebraic topology, unsupervised learning'
     )
